from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain.chains import create_history_aware_retriever, create_retrieval_chain
from langchain.chains.combine_documents import create_stuff_documents_chain
from langchain_core.messages import AIMessage
from langchain_ollama import OllamaLLM

class Chatbot_OpenSource:
    def __init__(self, 
                 vectorstore, 
                 k=10, 
                 score_threshold=0.7, 
                 model_name='llama3.1', #Can be one of the hosted models on https://chat.test.dias.rm.dk/
                 temperature=0.0, 
                 chat_history=[AIMessage(content="Hej. Jeg er en AI-bot. Hvordan kan jeg hjælpe dig?")],
                 system_prompt=ChatPromptTemplate.from_messages([
        """
        "system", "Du er en chatbot, hjælper med at svare på spørgsmål om den kontekst, du er givet. Besvar brugerens spørgsmål baseret på nedenstående kontekst. 
        
        Du oplyser altid, alle de steder i konteksten, du har fundet svaret med "Kilde: ", indsætter den på en linje nedenfor.  

        {context}

        Du giver altid det svar i konteksten, der bedst kan besvare spørgsmålet, men er du ikke sikker på svaret, siger du altid, at du ikke er sikker.

        Du svarer ADLRIG på spørgsmål, du ikke kan finde en kilde på eller svar på i konteksten.
        """,
        MessagesPlaceholder(variable_name="chat_history"),
        ("user", "{input}"),
                                                                ])
    ):      
        
        #supported_models=['llama3', 'mistral', 'llama3:70b']
        
        #if model_name not in supported_models:
        #    raise ValueError(f"model_name must be one of the following supported models {supported_models}")
        
        self.vectorstore = vectorstore
        self.k = k
        self.score_threshold = score_threshold
        self.llm = OllamaLLM(temperature=temperature, base_url='https://ollama.ml-test.dias.rm.dk', model=model_name) #DIAS hoster modellen
        self.chat_history= chat_history
        self.system_prompt = system_prompt

    def get_context_retriever_chain(self):
        try:
            llm = self.llm

            retriever = self.vectorstore.as_retriever(
                search_type="similarity_score_threshold", 
                search_kwargs={"k": self.k, "score_threshold": self.score_threshold}
            )

            prompt = ChatPromptTemplate.from_messages([
                MessagesPlaceholder(variable_name="chat_history"),
                ("user", "{input}"),
                ("user", "På baggrund af ovenstående samtale, generer en søgeforespørgsel til at slå op i, for at få oplysninger der er relevante for samtalen")
            ])

            retriever_chain = create_history_aware_retriever(llm, retriever, prompt)
            
            return retriever_chain
    
        except Exception as e:
            print("An error occurred while creating the context retriever chain:", e)
            return None

    def get_conversational_rag_chain(self, retriever_chain): 
        try:
            llm = self.llm
            
            prompt = self.system_prompt
            
            stuff_documents_chain = create_stuff_documents_chain(llm, prompt)
            
            return create_retrieval_chain(retriever_chain, stuff_documents_chain)
        
        except Exception as e:
            print("An error occurred while creating the conversational RAG chain:", e)
            return None
        
    def get_response(self, user_input):
        try:
            retriever_chain = self.get_context_retriever_chain()
            conversation_rag_chain = self.get_conversational_rag_chain(retriever_chain)
            response = conversation_rag_chain.invoke({
                "chat_history": self.chat_history,
                "input": user_input
            })
            return response['answer']

        except Exception as e:
            print("An error occurred during response generation:", e)
            return "An error occurred, and we couldn't process your request."        